@@ -1,6 +1,6 @@
 /*
  * The 3D Studio File Format Library
- * Copyright (C) 1996-2001 by J.E. Hoffmann <je-h@gmx.net>
+ * Copyright (C) 1996-2007 by Jan Eric Kyprianidis <www.kyprianidis.com>
  * All rights reserved.
  *
  * This program is  free  software;  you can redistribute it and/or modify it
@@ -17,9 +17,8 @@
  * along with  this program;  if not, write to the  Free Software Foundation,
  * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: file.c,v 1.19 2001/08/07 10:25:43 vividos Exp $
+ * $Id: file.c,v 1.34 2007/06/20 17:04:08 jeh Exp $
  */
-#define LIB3DS_EXPORT
 #include <lib3ds/file.h>
 #include <lib3ds/chunk.h>
 #include <lib3ds/io.h>
@@ -28,20 +27,16 @@
 #include <lib3ds/camera.h>
 #include <lib3ds/light.h>
 #include <lib3ds/node.h>
+#include <lib3ds/matrix.h>
 #include <lib3ds/vector.h>
 #include <stdlib.h>
 #include <string.h>
 #include <math.h>
-#include <config.h>
-#ifdef WITH_DMALLOC
-#include <dmalloc.h>
-#endif
+#include <float.h>
 
 
 /*!
  * \defgroup file Files
- *
- * \author J.E. Hoffmann <je-h@gmx.net>
  */
 
 
@@ -84,16 +79,16 @@ fileio_tell_func(void *self)
 }
 
 
-static int
-fileio_read_func(void *self, Lib3dsByte *buffer, int size)
+static size_t
+fileio_read_func(void *self, void *buffer, size_t size)
 {
   FILE *f = (FILE*)self;
   return(fread(buffer, 1, size, f));
 }
 
 
-static int
-fileio_write_func(void *self, const Lib3dsByte *buffer, int size)
+static size_t
+fileio_write_func(void *self, const void *buffer, size_t size)
 {
   FILE *f = (FILE*)self;
   return(fwrite(buffer, 1, size, f));
@@ -109,7 +104,7 @@ fileio_write_func(void *self, const Lib3dsByte *buffer, int size)
  *           data of the .3DS file. 
  *           If the .3DS file can not be loaded NULL is returned.
  *
- * \note     To free the returned pointer use lib3ds_free.
+ * \note     To free the returned structure use lib3ds_free.
  *
  * \see lib3ds_file_save
  * \see lib3ds_file_new
@@ -124,7 +119,6 @@ lib3ds_file_load(const char *filename)
   Lib3dsFile *file;
   Lib3dsIo *io;
 
-
   f = fopen(filename, "rb");
   if (!f) {
     return(0);
@@ -134,6 +128,7 @@ lib3ds_file_load(const char *filename)
     fclose(f);
     return(0);
   }
+  
   io = lib3ds_io_new(
     f, 
     fileio_error_func,
@@ -150,6 +145,7 @@ lib3ds_file_load(const char *filename)
 
   if (!lib3ds_file_read(file, io)) {
     free(file);
+    lib3ds_io_free(io);
     fclose(f);
     return(0);
   }
@@ -207,6 +203,11 @@ lib3ds_file_save(Lib3dsFile *file, const char *filename)
 
 
 /*!
+ * Creates and returns a new, empty Lib3dsFile object.
+ *
+ * \return	A pointer to the Lib3dsFile structure.
+ *		If the structure cannot be allocated, NULL is returned.
+ *
  * \ingroup file
  */
 Lib3dsFile*
@@ -233,6 +234,10 @@ lib3ds_file_new()
 
 
 /*!
+ * Free a Lib3dsFile object and all of its resources.
+ *
+ * \param file The Lib3dsFile object to be freed.
+ *
  * \ingroup file
  */
 void
@@ -240,6 +245,7 @@ lib3ds_file_free(Lib3dsFile* file)
 {
   ASSERT(file);
   lib3ds_viewport_set_views(&file->viewport,0);
+  lib3ds_viewport_set_views(&file->viewport_keyf,0);
   {
     Lib3dsMaterial *p,*q;
     
@@ -289,6 +295,13 @@ lib3ds_file_free(Lib3dsFile* file)
 
 
 /*!
+ * Evaluate all of the nodes in this Lib3dsFile object.
+ *
+ * \param file The Lib3dsFile object to be evaluated.
+ * \param t time value, between 0. and file->frames
+ *
+ * \see lib3ds_node_eval
+ *
  * \ingroup file
  */
 void
@@ -308,6 +321,10 @@ named_object_read(Lib3dsFile *file, Lib3dsIo *io)
   Lib3dsChunk c;
   char name[64];
   Lib3dsWord chunk;
+  Lib3dsMesh *mesh = NULL;
+  Lib3dsCamera *camera = NULL;
+  Lib3dsLight *light = NULL;
+  Lib3dsDword object_flags;
 
   if (!lib3ds_chunk_read_start(&c, LIB3DS_NAMED_OBJECT, io)) {
     return(LIB3DS_FALSE);
@@ -318,12 +335,11 @@ named_object_read(Lib3dsFile *file, Lib3dsIo *io)
   lib3ds_chunk_dump_info("  NAME=%s", name);
   lib3ds_chunk_read_tell(&c, io);
 
+  object_flags = 0;
   while ((chunk=lib3ds_chunk_read_next(&c, io))!=0) {
     switch (chunk) {
       case LIB3DS_N_TRI_OBJECT:
         {
-          Lib3dsMesh *mesh;
-
           mesh=lib3ds_mesh_new(name);
           if (!mesh) {
             return(LIB3DS_FALSE);
@@ -335,10 +351,9 @@ named_object_read(Lib3dsFile *file, Lib3dsIo *io)
           lib3ds_file_insert_mesh(file, mesh);
         }
         break;
+      
       case LIB3DS_N_CAMERA:
         {
-          Lib3dsCamera *camera;
-
           camera=lib3ds_camera_new(name);
           if (!camera) {
             return(LIB3DS_FALSE);
@@ -350,10 +365,9 @@ named_object_read(Lib3dsFile *file, Lib3dsIo *io)
           lib3ds_file_insert_camera(file, camera);
         }
         break;
+      
       case LIB3DS_N_DIRECT_LIGHT:
         {
-          Lib3dsLight *light;
-
           light=lib3ds_light_new(name);
           if (!light) {
             return(LIB3DS_FALSE);
@@ -365,10 +379,46 @@ named_object_read(Lib3dsFile *file, Lib3dsIo *io)
           lib3ds_file_insert_light(file, light);
         }
         break;
+      
+      case LIB3DS_OBJ_HIDDEN:
+        object_flags |= LIB3DS_OBJECT_HIDDEN;
+        break;
+
+      case LIB3DS_OBJ_DOESNT_CAST:
+        object_flags |= LIB3DS_OBJECT_DOESNT_CAST;
+        break;
+
+      case LIB3DS_OBJ_VIS_LOFTER:
+        object_flags |= LIB3DS_OBJECT_VIS_LOFTER;
+        break;
+
+      case LIB3DS_OBJ_MATTE:
+        object_flags |= LIB3DS_OBJECT_MATTE;
+        break;
+
+      case LIB3DS_OBJ_DONT_RCVSHADOW:
+        object_flags |= LIB3DS_OBJECT_DONT_RCVSHADOW;
+        break;
+
+      case LIB3DS_OBJ_FAST:
+        object_flags |= LIB3DS_OBJECT_FAST;
+        break;
+
+      case LIB3DS_OBJ_FROZEN:
+        object_flags |= LIB3DS_OBJECT_FROZEN;
+        break;
+
       default:
         lib3ds_chunk_unknown(chunk);
     }
   }
+
+  if (mesh)
+    mesh->object_flags = object_flags;
+  if (camera)
+    camera->object_flags = object_flags;
+  if (light)
+    light->object_flags = object_flags;
   
   lib3ds_chunk_read_end(&c, io);
   return(LIB3DS_TRUE);
@@ -544,6 +594,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
 {
   Lib3dsChunk c;
   Lib3dsWord chunk;
+  Lib3dsDword node_number = 0;
 
   if (!lib3ds_chunk_read_start(&c, LIB3DS_KFDATA, io)) {
     return(LIB3DS_FALSE);
@@ -588,6 +639,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -603,6 +655,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -618,6 +671,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -633,6 +687,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -649,6 +704,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -664,6 +720,7 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
           if (!node) {
             return(LIB3DS_FALSE);
           }
+          node->node_id=node_number++;
           lib3ds_chunk_read_reset(&c, io);
           if (!lib3ds_node_read(node, file, io)) {
             return(LIB3DS_FALSE);
@@ -682,6 +739,13 @@ kfdata_read(Lib3dsFile *file, Lib3dsIo *io)
 
 
 /*!
+ * Read 3ds file data into a Lib3dsFile object.
+ *
+ * \param file The Lib3dsFile object to be filled.
+ * \param io A Lib3dsIo object previously set up by the caller.
+ *
+ * \return LIB3DS_TRUE on success, LIB3DS_FALSE on failure.
+ *
  * \ingroup file
  */
 Lib3dsBool
@@ -763,6 +827,43 @@ colorf_write(Lib3dsRgba rgb, Lib3dsIo *io)
 }
 
 
+static Lib3dsBool
+object_flags_write(Lib3dsDword flags, Lib3dsIo *io)
+{
+  if (flags){
+    if (flags & LIB3DS_OBJECT_HIDDEN) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_HIDDEN, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_VIS_LOFTER) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_VIS_LOFTER, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_DOESNT_CAST) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_DOESNT_CAST, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_MATTE) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_MATTE, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_DONT_RCVSHADOW) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_DOESNT_CAST, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_FAST) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_FAST, io))
+        return LIB3DS_FALSE;
+    }
+    if (flags & LIB3DS_OBJECT_FROZEN) {
+      if (!lib3ds_chunk_write_switch(LIB3DS_OBJ_FROZEN, io))
+        return LIB3DS_FALSE;
+    }
+  }
+  return LIB3DS_TRUE;
+}
+
+
 static Lib3dsBool
 mdata_write(Lib3dsFile *file, Lib3dsIo *io)
 {
@@ -841,6 +942,7 @@ mdata_write(Lib3dsFile *file, Lib3dsIo *io)
       }
       lib3ds_io_write_string(io, p->name);
       lib3ds_camera_write(p,io);
+      object_flags_write(p->object_flags,io);
       if (!lib3ds_chunk_write_end(&c,io)) {
         return(LIB3DS_FALSE);
       }
@@ -857,6 +959,7 @@ mdata_write(Lib3dsFile *file, Lib3dsIo *io)
       }
       lib3ds_io_write_string(io,p->name);
       lib3ds_light_write(p,io);
+      object_flags_write(p->object_flags,io);
       if (!lib3ds_chunk_write_end(&c,io)) {
         return(LIB3DS_FALSE);
       }
@@ -873,6 +976,7 @@ mdata_write(Lib3dsFile *file, Lib3dsIo *io)
       }
       lib3ds_io_write_string(io, p->name);
       lib3ds_mesh_write(p,io);
+      object_flags_write(p->object_flags,io);
       if (!lib3ds_chunk_write_end(&c,io)) {
         return(LIB3DS_FALSE);
       }
@@ -907,6 +1011,10 @@ static Lib3dsBool
 kfdata_write(Lib3dsFile *file, Lib3dsIo *io)
 {
   Lib3dsChunk c;
+
+  if (!file->nodes) {
+    return(LIB3DS_TRUE);
+  }
   
   c.chunk=LIB3DS_KFDATA;
   if (!lib3ds_chunk_write_start(&c,io)) {
@@ -916,7 +1024,7 @@ kfdata_write(Lib3dsFile *file, Lib3dsIo *io)
   { /*---- LIB3DS_KFHDR ----*/
     Lib3dsChunk c;
     c.chunk=LIB3DS_KFHDR;
-    c.size=6 + 2 + strlen(file->name)+1 +4;
+    c.size=6 + 2 + (Lib3dsDword)strlen(file->name)+1 +4;
     lib3ds_chunk_write(&c,io);
     lib3ds_io_write_intw(io, file->keyf_revision);
     lib3ds_io_write_string(io, file->name);
@@ -959,6 +1067,13 @@ kfdata_write(Lib3dsFile *file, Lib3dsIo *io)
 
 
 /*!
+ * Write 3ds file data from a Lib3dsFile object to a file.
+ *
+ * \param file The Lib3dsFile object to be written.
+ * \param io A Lib3dsIo object previously set up by the caller.
+ *
+ * \return LIB3DS_TRUE on success, LIB3DS_FALSE on failure.
+ *
  * \ingroup file
  */
 Lib3dsBool
@@ -996,6 +1111,15 @@ lib3ds_file_write(Lib3dsFile *file, Lib3dsIo *io)
 
 
 /*!
+ * Insert a new Lib3dsMaterial object into the materials list of
+ * a Lib3dsFile object.
+ *
+ * The new Lib3dsMaterial object is inserted into the materials list
+ * in alphabetic order by name.
+ *
+ * \param file The Lib3dsFile object to be modified.
+ * \param material The Lib3dsMaterial object to be inserted into file->materials
+ *
  * \ingroup file
  */
 void
@@ -1026,6 +1150,15 @@ lib3ds_file_insert_material(Lib3dsFile *file, Lib3dsMaterial *material)
 
 
 /*!
+ * Remove a Lib3dsMaterial object from the materials list of
+ * a Lib3dsFile object.
+ *
+ * If the Lib3dsMaterial is not found in the materials list, nothing is
+ * done (except that an error log message may be generated.)
+ *
+ * \param file The Lib3dsFile object to be modified.
+ * \param material The Lib3dsMaterial object to be removed from file->materials
+ *
  * \ingroup file
  */
 void
@@ -1056,6 +1189,13 @@ lib3ds_file_remove_material(Lib3dsFile *file, Lib3dsMaterial *material)
 
 
 /*!
+ * Return a Lib3dsMaterial object by name.
+ *
+ * \param file Lib3dsFile object to be searched.
+ * \param name Name of the Lib3dsMaterial object to be searched for.
+ *
+ * \return A pointer to the named Lib3dsMaterial, or NULL if not found.
+ *
  * \ingroup file
  */
 Lib3dsMaterial*
@@ -1074,6 +1214,12 @@ lib3ds_file_material_by_name(Lib3dsFile *file, const char *name)
 
 
 /*!
+ * Dump all Lib3dsMaterial objects found in a Lib3dsFile object.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_material_dump
+ *
  * \ingroup file
  */
 void
@@ -1089,6 +1235,15 @@ lib3ds_file_dump_materials(Lib3dsFile *file)
 
 
 /*!
+ * Insert a new Lib3dsMesh object into the meshes list of
+ * a Lib3dsFile object.
+ *
+ * The new Lib3dsMesh object is inserted into the meshes list
+ * in alphabetic order by name.
+ *
+ * \param file  The Lib3dsFile object to be modified.
+ * \param mesh  The Lib3dsMesh object to be inserted into file->meshes
+ *
  * \ingroup file
  */
 void
@@ -1119,6 +1274,15 @@ lib3ds_file_insert_mesh(Lib3dsFile *file, Lib3dsMesh *mesh)
 
 
 /*!
+ * Remove a Lib3dsMesh object from the meshes list of
+ * a Lib3dsFile object.
+ *
+ * If the Lib3dsMesh is not found in the meshes list, nothing is done
+ * (except that an error log message may be generated.)
+ *
+ * \param file  The Lib3dsFile object to be modified.
+ * \param mesh  The Lib3dsMesh object to be removed from file->meshes
+ *
  * \ingroup file
  */
 void
@@ -1149,6 +1313,13 @@ lib3ds_file_remove_mesh(Lib3dsFile *file, Lib3dsMesh *mesh)
 
 
 /*!
+ * Return a Lib3dsMesh object from a Lib3dsFile by name.
+ *
+ * \param file Lib3dsFile object to be searched.
+ * \param name Name of the Lib3dsMesh object to be searched for.
+ *
+ * \return A pointer to the named Lib3dsMesh, or NULL if not found.
+ *
  * \ingroup file
  */
 Lib3dsMesh*
@@ -1167,6 +1338,12 @@ lib3ds_file_mesh_by_name(Lib3dsFile *file, const char *name)
 
 
 /*!
+ * Dump all Lib3dsMesh objects found in a Lib3dsFile object.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_mesh_dump
+ *
  * \ingroup file
  */
 void
@@ -1202,6 +1379,18 @@ dump_instances(Lib3dsNode *node, const char* parent)
 
 
 /*!
+ * Dump all Lib3dsNode object names found in a Lib3dsFile object.
+ *
+ * For each node of type OBJECT_NODE, its name and data.object.instance
+ * fields are printed to stdout.  Consider using lib3ds_file_dump_nodes()
+ * instead, as that function dumps more information.
+ *
+ * Nodes are dumped recursively.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_file_dump_nodes
+ *
  * \ingroup file
  */
 void
@@ -1217,6 +1406,15 @@ lib3ds_file_dump_instances(Lib3dsFile *file)
 
 
 /*!
+ * Insert a new Lib3dsCamera object into the cameras list of
+ * a Lib3dsFile object.
+ *
+ * The new Lib3dsCamera object is inserted into the cameras list
+ * in alphabetic order by name.
+ *
+ * \param file      The Lib3dsFile object to be modified.
+ * \param camera    The Lib3dsCamera object to be inserted into file->cameras
+ *
  * \ingroup file
  */
 void
@@ -1247,6 +1445,15 @@ lib3ds_file_insert_camera(Lib3dsFile *file, Lib3dsCamera *camera)
 
 
 /*!
+ * Remove a Lib3dsCamera object from the cameras list of
+ * a Lib3dsFile object.
+ *
+ * If the Lib3dsCamera is not found in the cameras list, nothing is done
+ * (except that an error log message may be generated.)
+ *
+ * \param file      The Lib3dsFile object to be modified.
+ * \param camera    The Lib3dsCamera object to be removed from file->cameras
+ *
  * \ingroup file
  */
 void
@@ -1277,6 +1484,13 @@ lib3ds_file_remove_camera(Lib3dsFile *file, Lib3dsCamera *camera)
 
 
 /*!
+ * Return a Lib3dsCamera object from a Lib3dsFile by name.
+ *
+ * \param file Lib3dsFile object to be searched.
+ * \param name Name of the Lib3dsCamera object to be searched for.
+ *
+ * \return A pointer to the named Lib3dsCamera, or NULL if not found.
+ *
  * \ingroup file
  */
 Lib3dsCamera*
@@ -1295,6 +1509,12 @@ lib3ds_file_camera_by_name(Lib3dsFile *file, const char *name)
 
 
 /*!
+ * Dump all Lib3dsCamera objects found in a Lib3dsFile object.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_camera_dump
+ *
  * \ingroup file
  */
 void
@@ -1310,6 +1530,15 @@ lib3ds_file_dump_cameras(Lib3dsFile *file)
 
 
 /*!
+ * Insert a new Lib3dsLight object into the lights list of
+ * a Lib3dsFile object.
+ *
+ * The new Lib3dsLight object is inserted into the lights list
+ * in alphabetic order by name.
+ *
+ * \param file  The Lib3dsFile object to be modified.
+ * \param light The Lib3dsLight object to be inserted into file->lights
+ *
  * \ingroup file
  */
 void
@@ -1340,6 +1569,15 @@ lib3ds_file_insert_light(Lib3dsFile *file, Lib3dsLight *light)
 
 
 /*!
+ * Remove a Lib3dsLight object from the lights list of
+ * a Lib3dsFile object.
+ *
+ * If the Lib3dsLight is not found in the lights list, nothing is done
+ * (except that an error log message may be generated.)
+ *
+ * \param file  The Lib3dsFile object to be modified.
+ * \param light The Lib3dsLight object to be removed from file->lights
+ *
  * \ingroup file
  */
 void
@@ -1370,6 +1608,13 @@ lib3ds_file_remove_light(Lib3dsFile *file, Lib3dsLight *light)
 
 
 /*!
+ * Return a Lib3dsLight object from a Lib3dsFile by name.
+ *
+ * \param file Lib3dsFile object to be searched.
+ * \param name Name of the Lib3dsLight object to be searched for.
+ *
+ * \return A pointer to the named Lib3dsLight, or NULL if not found.
+ *
  * \ingroup file
  */
 Lib3dsLight*
@@ -1388,6 +1633,12 @@ lib3ds_file_light_by_name(Lib3dsFile *file, const char *name)
 
 
 /*!
+ * Dump all Lib3dsLight objects found in a Lib3dsFile object.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_light_dump
+ *
  * \ingroup file
  */
 void
@@ -1403,67 +1654,19 @@ lib3ds_file_dump_lights(Lib3dsFile *file)
 
 
 /*!
- * \ingroup file
- */
-void
-lib3ds_file_bounding_box(Lib3dsFile *file, Lib3dsVector min, Lib3dsVector max)
-{
-  Lib3dsBool init=LIB3DS_FALSE;
-
-  {
-    Lib3dsVector lmin, lmax;
-    Lib3dsMesh *p=file->meshes;
-
-    if (!init && p) {
-      init = LIB3DS_TRUE;
-      lib3ds_mesh_bounding_box(p, min, max);
-      p = p->next;  
-    }
-    while (p) {
-      lib3ds_mesh_bounding_box(p, lmin, lmax);
-      lib3ds_vector_min(min, lmin);
-      lib3ds_vector_max(max, lmax);
-      p=p->next;
-    }
-  }
-  {
-    Lib3dsCamera *p=file->cameras;
-    if (!init && p) {
-      init = LIB3DS_TRUE;
-      lib3ds_vector_copy(min, p->position);
-      lib3ds_vector_copy(max, p->position);
-    }
-
-    while (p) {
-      lib3ds_vector_min(min, p->position);
-      lib3ds_vector_max(max, p->position);
-      lib3ds_vector_min(min, p->target);
-      lib3ds_vector_max(max, p->target);
-      p=p->next;
-    }
-  }
-  {
-    Lib3dsLight *p=file->lights;
-    if (!init && p) {
-      init = LIB3DS_TRUE;
-      lib3ds_vector_copy(min, p->position);
-      lib3ds_vector_copy(max, p->position);
-    }
-
-    while (p) {
-      lib3ds_vector_min(min, p->position);
-      lib3ds_vector_max(max, p->position);
-      if (p->spot_light) {
-        lib3ds_vector_min(min, p->spot);
-        lib3ds_vector_max(max, p->spot);
-      }
-      p=p->next;
-    }
-  }
-}  
-
-
-/*!
+ * Return a node object by name and type.
+ *
+ * This function performs a recursive search for the specified node.
+ * Both name and type must match.
+ *
+ * \param file The Lib3dsFile to be searched.
+ * \param name The target node name.
+ * \param type The target node type
+ *
+ * \return A pointer to the first matching node, or NULL if not found.
+ *
+ * \see lib3ds_node_by_name
+ *
  * \ingroup file
  */
 Lib3dsNode*
@@ -1486,6 +1689,17 @@ lib3ds_file_node_by_name(Lib3dsFile *file, const char* name, Lib3dsNodeTypes typ
 
 
 /*!
+ * Return a node object by id.
+ *
+ * This function performs a recursive search for the specified node.
+ *
+ * \param file The Lib3dsFile to be searched.
+ * \param node_id The target node id.
+ *
+ * \return A pointer to the first matching node, or NULL if not found.
+ *
+ * \see lib3ds_node_by_id
+ *
  * \ingroup file
  */
 Lib3dsNode*
@@ -1508,6 +1722,22 @@ lib3ds_file_node_by_id(Lib3dsFile *file, Lib3dsWord node_id)
 
 
 /*!
+ * Insert a new node into a Lib3dsFile object.
+ *
+ * If the node's parent_id structure is not LIB3DS_NO_PARENT and the
+ * specified parent is found inside the Lib3dsFile object, then the
+ * node is inserted as a child of that parent.  If the parent_id
+ * structure is LIB3DS_NO_PARENT or the specified parent is not found,
+ * then the node is inserted at the top level.
+ *
+ * Node is inserted in alphabetic order by name.
+ *
+ * Finally, if any other top-level nodes in file specify this node as
+ * their parent, they are relocated as a child of this node.
+ *
+ * \param file The Lib3dsFile object to be modified.
+ * \param node The node to be inserted into file
+ *
  * \ingroup file
  */
 void
@@ -1569,6 +1799,13 @@ lib3ds_file_insert_node(Lib3dsFile *file, Lib3dsNode *node)
 
 
 /*!
+ * Remove a node from the a Lib3dsFile object.
+ *
+ * \param file The Lib3dsFile object to be modified.
+ * \param node The Lib3dsNode object to be removed from file
+ *
+ * \return LIB3DS_TRUE on success, LIB3DS_FALSE if node is not found in file
+ *
  * \ingroup file
  */
 Lib3dsBool
@@ -1615,6 +1852,167 @@ lib3ds_file_remove_node(Lib3dsFile *file, Lib3dsNode *node)
 
 
 /*!
+ * This function computes the bounding box of meshes, cameras and lights 
+ * defined in the 3D editor.
+ *
+ * \param file              The Lib3dsFile object to be examined.
+ * \param include_meshes    Include meshes in bounding box calculation.
+ * \param include_cameras   Include cameras in bounding box calculation.
+ * \param include_lights    Include lights in bounding box calculation.
+ * \param bmin              Returned minimum x,y,z values.
+ * \param bmax              Returned maximum x,y,z values.
+ *
+ * \ingroup file
+ */
+void
+lib3ds_file_bounding_box_of_objects(Lib3dsFile *file, Lib3dsBool include_meshes, 
+                                    Lib3dsBool include_cameras, Lib3dsBool include_lights, 
+                                    Lib3dsVector bmin, Lib3dsVector bmax)
+{
+  bmin[0] = bmin[1] = bmin[2] = FLT_MAX; 
+  bmax[0] = bmax[1] = bmax[2] = FLT_MIN; 
+
+  if (include_meshes) {
+    Lib3dsVector lmin, lmax;
+    Lib3dsMesh *p=file->meshes;
+    while (p) {
+      lib3ds_mesh_bounding_box(p, lmin, lmax);
+      lib3ds_vector_min(bmin, lmin);
+      lib3ds_vector_max(bmax, lmax);
+      p=p->next;
+    }
+  }
+  if (include_cameras) {
+    Lib3dsCamera *p=file->cameras;
+    while (p) {
+      lib3ds_vector_min(bmin, p->position);
+      lib3ds_vector_max(bmax, p->position);
+      lib3ds_vector_min(bmin, p->target);
+      lib3ds_vector_max(bmax, p->target);
+      p=p->next;
+    }
+  }
+  if (include_lights) {
+    Lib3dsLight *p=file->lights;
+    while (p) {
+      lib3ds_vector_min(bmin, p->position);
+      lib3ds_vector_max(bmax, p->position);
+      if (p->spot_light) {
+        lib3ds_vector_min(bmin, p->spot);
+        lib3ds_vector_max(bmax, p->spot);
+      }
+      p=p->next;
+    }
+  }
+}  
+
+
+static void
+file_bounding_box_of_nodes_impl(Lib3dsNode *node, Lib3dsFile *file, Lib3dsBool include_meshes, 
+                                Lib3dsBool include_cameras, Lib3dsBool include_lights, 
+                                Lib3dsVector bmin, Lib3dsVector bmax)
+{
+  switch (node->type)
+  {
+    case LIB3DS_OBJECT_NODE:
+      if (include_meshes) {
+        Lib3dsMesh *mesh;
+
+        mesh = lib3ds_file_mesh_by_name(file, node->data.object.instance);
+        if (!mesh)
+          mesh = lib3ds_file_mesh_by_name(file, node->name);
+        if (mesh) {
+          Lib3dsMatrix inv_matrix, M;
+          Lib3dsVector v;
+          unsigned i;
+
+          lib3ds_matrix_copy(inv_matrix, mesh->matrix);
+          lib3ds_matrix_inv(inv_matrix);
+          lib3ds_matrix_copy(M, node->matrix);
+          lib3ds_matrix_translate_xyz(M, -node->data.object.pivot[0], -node->data.object.pivot[1], -node->data.object.pivot[2]);
+          lib3ds_matrix_mult(M, inv_matrix);
+
+          for (i=0; i<mesh->points; ++i) {
+            lib3ds_vector_transform(v, M, mesh->pointL[i].pos);
+            lib3ds_vector_min(bmin, v);
+            lib3ds_vector_max(bmax, v);
+          }
+        }
+      }
+      break;
+   /*
+    case LIB3DS_CAMERA_NODE:
+    case LIB3DS_TARGET_NODE:
+      if (include_cameras) {
+        Lib3dsVector z,v;
+        lib3ds_vector_zero(z);
+        lib3ds_vector_transform(v, node->matrix, z);
+        lib3ds_vector_min(bmin, v);
+        lib3ds_vector_max(bmax, v);
+      }
+      break;
+
+    case LIB3DS_LIGHT_NODE:
+    case LIB3DS_SPOT_NODE:
+      if (include_lights) {
+        Lib3dsVector z,v;
+        lib3ds_vector_zero(z);
+        lib3ds_vector_transform(v, node->matrix, z);
+        lib3ds_vector_min(bmin, v);
+        lib3ds_vector_max(bmax, v);
+      }
+      break;
+    */
+  }
+  {
+    Lib3dsNode *p=node->childs;
+    while (p) {
+      file_bounding_box_of_nodes_impl(p, file, include_meshes, include_cameras, include_lights, bmin, bmax);
+      p=p->next;
+    }
+  }
+}
+
+
+/*!
+ * This function computes the bounding box of mesh, camera and light instances 
+ * defined in the Keyframer.
+ *
+ * \param file              The Lib3dsFile object to be examined.
+ * \param include_meshes    Include meshes in bounding box calculation.
+ * \param include_cameras   Include cameras in bounding box calculation.
+ * \param include_lights    Include lights in bounding box calculation.
+ * \param bmin              Returned minimum x,y,z values.
+ * \param bmax              Returned maximum x,y,z values.
+ *
+ * \ingroup file
+ */
+void
+lib3ds_file_bounding_box_of_nodes(Lib3dsFile *file, Lib3dsBool include_meshes, 
+                                  Lib3dsBool include_cameras, Lib3dsBool include_lights, 
+                                  Lib3dsVector bmin, Lib3dsVector bmax)
+{
+  Lib3dsNode *p;
+  
+  bmin[0] = bmin[1] = bmin[2] = FLT_MAX; 
+  bmax[0] = bmax[1] = bmax[2] = FLT_MIN; 
+  p=file->nodes;
+  while (p) {
+    file_bounding_box_of_nodes_impl(p, file, include_meshes, include_cameras, include_lights, bmin, bmax);
+    p=p->next;
+  }
+}
+
+
+/*!
+ * Dump all node objects found in a Lib3dsFile object.
+ *
+ * Nodes are dumped recursively.
+ *
+ * \param file Lib3dsFile object to be dumped.
+ *
+ * \see lib3ds_node_dump
+ *
  * \ingroup file
  */
 void
@@ -1628,11 +2026,3 @@ lib3ds_file_dump_nodes(Lib3dsFile *file)
   }
 }
 
-
-/*!
-
-\typedef Lib3dsFile
-  \ingroup file
-  \sa _Lib3dsFile
-
-*/
